/* ----------------------------------------------------------------------------
Copyright (c) 2018-2026 Microsoft Research, Daan Leijen
This is free software; you can redistribute it and/or modify it under the
terms of the MIT license.
-----------------------------------------------------------------------------*/

#define TEST_STRESS_SUBPROCS 1
#define NSUBPROCS            2
#define NTHREADS             16
#include "test-stress.c"
